@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 import argparse
-import xasyVersion
+from xasyversion.version import VERSION as xasyVersion
 import PyQt5.QtCore as QtCore
 # Add arguments here.
 
@@ -11,7 +11,7 @@ def parseArgs(args):
                         help='Custom path to asy executable')
     parser.add_argument('-v', '-version', '--version',
                         help='Version number', action='version',
-                        version='xasy v{0}'.format(xasyVersion.xasyVersion))
+                        version='xasy v{0}'.format(xasyVersion))
     parser.add_argument('-l', '-language', '--language',
                         help='language')
     parser.add_argument('-x', '-mag', '--mag',
@@ -20,6 +20,11 @@ def parseArgs(args):
     parser.add_argument('-render', '--render',
                         help='Number of pixels per bp in 3D rendered bitmaps',
                         default=None, type=float)
+    parser.add_argument('-additional-asy-args', '--additional-asy-args',
+                        help='Comma-separated values of additional arguments to pass'
+                             'to Asymptote',
+                        dest='additionalAsyArgs',
+                        type=str, default=None)
     parser.add_argument(
         'filename',
         help='Filename to load (if omitted, initialize blank canvas)',
