@@ -13,6 +13,56 @@
  * in env.h.
  *****/
 
+/* Implementation of templated modules:
+ *
+ *    Translating an access declaration:
+ *
+ *    access Map(Key=A, Value=B) as MapAB;
+ *
+ *    run encodeLevel for both A and B
+ *    this should give the parent records for each struct
+ *    encode pushing the *number* of parents on the stack (i.e., push a
+ *        single int)
+ *    encode pushing the string "Map/1234567" on the stack
+ *    encode call to builtin loadTemplatedModule
+ *    also save into MapAB (varinit)
+ *
+ *    build list of types (or tyEntry?)
+ *
+ *    also ensure names match
+ *
+ *    *****
+ *
+ *    At runtime, loadTemplatedModule pops the string
+ *
+ *    if the module is already loaded, it pops the levels
+ *    and returns the already loaded module.
+ *
+ *    if the module is not loaded, it leaves the levels on the stack
+ *    and calls the initializer for the templated module
+ *
+ *    it might be easiest to give the number of pushed params as an argument
+ *    to loadTemplatedModule (ints and strings have no push/pop)
+ *
+ *    *****
+ *
+ *    Translating a templated module
+ *
+ *    we start translating a file with a list of (name, type) pairs
+ *
+ *    for each record type,
+ *    build variables for each parent level
+ *    and encode bytecode to pop the parents off the stack into these vars
+ *
+ *    build tyEntry for each templated type
+ *    if its a record, then use the above variables as ent->v
+ *
+ *    from here,
+ *    translate the file as a module as usual
+ *
+ *
+ */
+
 #include <sstream>
 
 #include <algorithm>
@@ -37,11 +87,6 @@ using namespace types;
 using settings::getSetting;
 using settings::Setting;
 
-// Dynamic loading of external libraries.
-types::record *transExternalModule(
-    trans::genv& ge, string filename, symbol id
-);
-
 namespace trans {
 
 genv::genv()
@@ -71,13 +116,6 @@ bool endswith(string suffix, string str)
 }
 
 record *genv::loadModule(symbol id, string filename) {
-  // Hackish way to load an external library.
-#if 0
-  if (endswith(".so", filename)) {
-    return transExternalModule(*this, filename, id);
-  }
-#endif
-
   // Get the abstract syntax tree.
   absyntax::file *ast = parser::parseFile(filename,"Loading");
 
@@ -95,15 +133,8 @@ record *genv::loadModule(symbol id, string filename) {
 record *genv::loadTemplatedModule(
       symbol id,
       string filename,
-      mem::vector<absyntax::namedTyEntry*> *args,
-      coenv& e
+      mem::vector<absyntax::namedTy*> *args
 ) {
-  // Hackish way to load an external library.
-#if 0
-  if (endswith(".so", filename)) {
-    return transExternalModule(*this, filename, id);
-  }
-#endif
 
   // Get the abstract syntax tree.
   absyntax::file *ast = parser::parseFile(filename,"Loading");
@@ -112,7 +143,7 @@ record *genv::loadTemplatedModule(
 
   em.sync();
 
-  record *r=ast->transAsTemplatedFile(*this, id, args, e);
+  record *r=ast->transAsTemplatedFile(*this, id, args);
 
   inTranslation.remove(filename);
 
@@ -138,7 +169,7 @@ record *genv::getModule(symbol id, string filename) {
   if (r)
     return r;
   else {
-    record *r=loadModule(id, filename);
+    r=loadModule(id, filename);
     // Don't add an erroneous module to the dictionary in interactive mode, as
     // the user may try to load it again.
     if (!interact::interactive || !em.errors()) {
@@ -150,18 +181,26 @@ record *genv::getModule(symbol id, string filename) {
 }
 
 record *genv::getTemplatedModule(
-    symbol index,
     string filename,
-    mem::vector<absyntax::namedTyEntry*>* args,
-    coenv& e
+    mem::vector<absyntax::namedTy*>* args
 ) {
   checkRecursion(filename);
 
+  types::signature* sig = new types::signature();
+  stringstream buf;
+  buf << filename << "/";
+  for (auto arg : *args) {
+    sig->add(formal(arg->t, arg->dest));
+    buf << arg->dest << "/";
+  }
+  buf << sig->handle() << "/";
+  symbol index=symbol::literalTrans(buf.str());
+
   record *r=imap[index];
   if (r)
     return r;
   else {
-    record *r=loadTemplatedModule(index, filename, args, e);
+    r=loadTemplatedModule(index, filename, args);
     // Don't add an erroneous module to the dictionary in interactive mode, as
     // the user may try to load it again.
     if (!interact::interactive || !em.errors()) {
